import * as React from 'react';

function _ref(prev, action) {
  return {
    // This is the current status of the promise or async/await function. A
    // promise or async/await can only be in one state at a time.
    status: action.status,
    // The value is persisted between 'success' statuses. This means I can
    // still display things that depend on my current value while my new
    // value is loading.
    value: action.status === 'success' ? action.value : prev.value,
    // Errors get reset each time we leave the error state. There's really
    // no use in keeping those around. They go stale once we leave.
    error: action.status === 'error' ? action.error : void 0,
    callback: prev.callback
  };
}

export const useAsync = asyncCallback => {
  // Creates a stable callback that manages our loading/success/error status updates
  // as the callback is invoked.
  const storedCallback = React.useRef(asyncCallback);
  storedCallback.current = asyncCallback;
  const [state, dispatch] = React.useReducer(_ref, void 0, () => {
    // This callback will not dispatch state once its been cancelled
    const cancelled = [];
    let running = storedCallback.current;

    const removeCancelled = current => cancelled.splice(cancelled.indexOf(current), 1);

    const callback = Object.assign(async (...args) => {
      dispatch({
        status: 'loading'
      });
      const current = running = storedCallback.current;
      removeCancelled(current);

      try {
        const value = await current(...args);
        cancelled.indexOf(current) === -1 && dispatch({
          status: 'success',
          value
        });
        removeCancelled(current);
      } catch (error) {
        cancelled.indexOf(current) === -1 && dispatch({
          status: 'error',
          error
        });
        removeCancelled(current);
      }
    }, {
      cancel: () => cancelled.push(running)
    });
    return {
      status: 'idle',
      value: void 0,
      error: void 0,
      callback
    };
  }); // Cancels any pending async callbacks when the hook unmounts

  React.useEffect(() => state.callback.cancel, [state.callback]);

  function _ref2() {
    // Prevent the callback from dispatching
    state.callback.cancel(); // Create a new callback and set status to cancelled

    dispatch({
      status: 'cancelled'
    });
  }

  return [React.useMemo(() => ({
    status: state.status,
    value: state.value,
    error: state.error,
    cancel: _ref2
  }), [state]), state.callback];
};
export const useAsyncEffect = (asyncCallback, dependencies) => {
  const [state, callback] = useAsync(asyncCallback); // Runs the callback each time we receive a new one

  React.useEffect(() => {
    callback(); // Uses callback.cancel rather than state.cancel so that the user never sees
    // a 'cancelled' status

    return callback.cancel; // eslint-disable-next-line react-hooks/exhaustive-deps
  }, dependencies);
  return state;
};